pub static PKGDATADIR: &str = "/usr/share/video-trimmer";
pub static LOCALEDIR: &str = "/usr/share/locale";
pub static VERSION: &str = "26.03.1";
pub static NAME_SUFFIX: &str = "";
pub static APP_ID: &str = "org.gnome.gitlab.YaLTeR.VideoTrimmer";
pub static G_LOG_DOMAIN: &str = "VideoTrimmer";
pub static PROFILE: &str = "";
